/*
  This is a basic helper function checking if the P5js sketch is loaded inside
  Max jweb object.
*/
function detectMax() {
  try {
    /*
      For references to all functions attached to window.max object read the
      "Communicating with Max from within jweb" document from Max documentation.
    */
    window.max.outlet('Hello Max!');
    return true;
  }
  catch(e) {
    console.log('Max, where are you?');
  }
  return false;
}

/*
  Here we are creating actual P5js sketch in the instance mode
  (look at https://github.com/processing/p5.js/wiki/p5.js-overview
  for details about P5js instantiation and namespace) to prevent
  binding P5js functions and variables to the "window" object. Thanks
  of that we can avoid namespaces conflict between Max and P5js.
*/
var s = function(p) {

  // An instance of the SpeachToText.
  var speechToText;
  var currentString = 'please say something';

  // let's test and memorize if the sketch is loaded inside Max jweb object
  var maxIsDetected = detectMax();

  /*
    This variable stores string identifies the name of the "channel" of data
    transmission through socketio.
  */
  var channelName = 'socketiobridge';

  // a few variables to store color data (just for this example sketch)
  var background_r = 255; var background_g = 255; var background_b = 255;
  var stroke_r = 0; var stroke_g = 0; var stroke_b = 0;

  function sttCallbackOnEndFunction(_v) {
    console.log("callbackOnEndFunction...");
    speechToText.recognizeSpeech();
  }
  
  function sttCallbackOnRecognizeFunction(_v) {
    if(_v === null)   return;
    if(_v.length < 1) return;
    if(_v === ' ')    return;
    _v = _v.trim();
    currentString = _v;
    /*
      If the motion detection system detected any spoken words, we send them
      through socket.io - theoretically, the detection of Max is unnecessary -
      I added it somewhat exaggerated.
    */
    if(!maxIsDetected) {
      try {
        socket.emit(channelName, {data: _v});
      } catch(e) {
        console.log('[sttCallbackOnRecognizeFunction] ' + e);
      }
    }
    console.log(_v);
  }

  /*
    Use windowResized function to adopt canvas size to the current size of
    the browser. It is particularly important when sketch is loaded inside
    the Max jweb object, which may be dynamically resized by the user.
  */
  p.windowResized = function() {
    p.resizeCanvas(innerWidth, innerHeight);
  }

  p.setup = function() {
    p.createCanvas(innerWidth, innerHeight);
    /*
      Basic Node/Socket.io init stuff.
    */
    socket = io.connect('http://127.0.0.1:8081');
    socket.on(
      channelName,
      function(_content) {
        switch(_content.status) {
          case 'connect':
            // that's great, nothing to do
          break;
          case 'message':
            if(!maxIsDetected) break;
            if(_content.data === null) {
              console.log('[tts] _content.data is null');
              break;
            }
            if(_content.data === undefined) {
              console.log('[tts] _content.data is undefined');
              break;
            }
            if(_content.data.length < 1) {
              console.log('[tts] _content.data.length < 1');
              break;
            }
            currentString = _content.data;
            window.max.outlet('stt', currentString);
          break;
          case 'disconnect':
            // that's great, nothing to do
          break;
          default: console.log('unhandled _content.status: ' + _content.status);
        }
      }
    );
    /*
      If the sketch is loaded inside Max jweb object we will carry out
      the necessary tasks to establish communitation between the patcher
      and the sketch (and to tune the sketch itself to present itself
      correctly when operating inside the jweb.
    */
    if(maxIsDetected) {
      // remove unwanted scroll bar
      document.getElementsByTagName('body')[0].style.overflow = 'hidden';
    }
    else {
      /*
        We assume that - because the website has not been loaded inside MaxMSP
        - we should start the speech detection system built into Google Chrome.
      */
      speechToText = new p.SpeechToText(
        p, sttCallbackOnRecognizeFunction, sttCallbackOnEndFunction
      );
      speechToText.recognizeSpeech();
    }
    p.noStroke();
    p.fill(255);
    p.textFont('Courier', 24);
    p.textAlign(p.CENTER, p.CENTER);
    p.frameRate(30);
  };

  p.draw = function() {
    if(!maxIsDetected) {
      if(speechToText.status < 0) {
        if(p.millis() - speechToText.recognizeCallTimeTag > 120000) {
          //console.log("reloading...");
          //location.reload();
          currentString = 'no speech recognized\n\nsay something, check the mic, reload the page';
        }
      }
    }
    p.background(0);
    p.text(currentString, p.width/2, p.height/2);
  };
};

// let's run the sketch in the "instance mode"
var myp5 = new p5(s);

/*
  SpeechToText - just as it sounds. This pseudo-class uses a speech detection
  engine built into the Google Chrome browser.
*/
p5.prototype.SpeechToText = function(
  _p5jsInstance,
  _callbackOnRecognizeFunction,
  _callbackOnEndFunction
) {
  this.p5jsInstance = _p5jsInstance;
  this.recognizeCallTimeTag = 0;
  this.final_transcript = '';
  this.recognizing = false;
  this.recognition;
  this.ignore_onend;
  this.start_timestamp;
  this.STATUS_FINISHED = -3;
  this.STATUS_NOT_SUPPORTED = -2;
  this.STATUS_ERROR = -1;
  this.STATUS_IDLE = 0;
  this.STATUS_BUSY = 1;
  this.status = this.STATUS_IDLE;
  if (!('webkitSpeechRecognition' in window)) {
    console.log("Hmmm... webkitSpeechRecognition not supported?");
    this.status = this.STATUS_NOT_SUPPORTED;
    return;
  }
  else {
    console.log("webkitSpeechRecognition detected");
  }
  this.recognition = new webkitSpeechRecognition();
  this.recognition.continuous = true;
  this.recognition.interimResults = true;
  this.recognition.lang = 'en-EN';
  
  this.recognition.onstart = function() {
    this.recognizing = true;
    console.log("webkitSpeechRecognition: start");
    this.status = this.STATUS_BUSY;
  };
  
  this.recognition.onerror = function(event) {
    this.status = this.STATUS_ERROR;
    console.log("webkitSpeechRecognition: event.error = " + event.error);
    if (event.error == 'no-speech') {
      console.log("webkitSpeechRecognition: event.error = no-speech");
      this.ignore_onend = true;
    }
    if (event.error == 'audio-capture') {
      console.log("webkitSpeechRecognition: event.error = audio-capture");
      this.ignore_onend = true;
    }
    if (event.error == 'not-allowed') {
      console.log("webkitSpeechRecognition: event.error = not-allowed");
      this.ignore_onend = true;
    }
    if (event.error == 'network') {
      console.log("webkitSpeechRecognition: event.error = network");
      //console.log("reloading...");
      //location.reload();
      this.ignore_onend = true;
    }
  };
  
  this.recognition.onend = function() {
    console.log("webkitSpeechRecognition: onend");
    _callbackOnEndFunction();
  };
  
  this.recognition.onresult = function(event) {
    var interim_transcript = '';
    if(typeof(event.results) == 'undefined') {
      this.recognition.onend = null;
      this.recognition.stop();
      this.status = this.STATUS_ERROR;
      return;
    }
    this.status = this.STATUS_BUSY;
    for(var i = event.resultIndex; i < event.results.length; ++i) {
      if(event.results[i].isFinal) {
        this.final_transcript += event.results[i][0].transcript;
      }
      else {
        interim_transcript += event.results[i][0].transcript;
      }
    }
    if (this.final_transcript || interim_transcript) {
      _callbackOnRecognizeFunction(interim_transcript);
    }
  };
}

p5.prototype.SpeechToText.prototype.recognizeSpeech = function() {
  console.log("(re)start speech recognition engine...");
  this.recognizeCallTimeTag = this.p5jsInstance.millis();
  this.recognition.start();
}